#!/usr/bin/env python3
from sum_dirac_dfcoef.args import args
from sum_dirac_dfcoef.file_writer import output_file_writer
from sum_dirac_dfcoef.functions_info import get_functions_info
from sum_dirac_dfcoef.header_info import HeaderInfo
from sum_dirac_dfcoef.privec_reader import PrivecProcessor
from sum_dirac_dfcoef.utils import get_dirac_filepath, should_write_electronic_results_to_file, should_write_positronic_results_to_file


def main() -> None:
    dirac_filepath = get_dirac_filepath()
    dirac_output = open(dirac_filepath, encoding="utf-8")
    dirac_output.seek(0)  # rewind to the beginning of the file
    header_info = HeaderInfo()
    if args.for_generator:
        header_info.read_header_info(dirac_output)
    dirac_output.seek(0)
    functions_info = get_functions_info(dirac_output)
    output_file_writer.create_blank_file()
    if not args.for_generator:
        # If the output file is not for dcaspt2_input_generator, don't write header information.
        output_file_writer.write_no_header_info()
    else:
        output_file_writer.write_headerinfo(header_info)

    # Read coefficients from the output file of DIRAC and store them in data_all_mo.
    privec_processor = PrivecProcessor(dirac_output, functions_info, header_info.eigenvalues)
    privec_processor.read_privec_data()
    data_all_mo = privec_processor.data_all_mo

    if should_write_electronic_results_to_file():
        add_blank_line = True if args.all_write else False
        output_file_writer.write_mo_data(data_all_mo.electronic, add_blank_line=add_blank_line)
    if should_write_positronic_results_to_file():
        output_file_writer.write_mo_data(data_all_mo.positronic, add_blank_line=False)



import copy
from enum import Enum, auto
from io import TextIOWrapper
from typing import Dict, List

from sum_dirac_dfcoef.args import args
from sum_dirac_dfcoef.atoms import AtomInfo
from sum_dirac_dfcoef.coefficient import get_coefficient
from sum_dirac_dfcoef.data import DataAllMO, DataMO
from sum_dirac_dfcoef.eigenvalues import Eigenvalues
from sum_dirac_dfcoef.functions_info import FunctionsInfo
from sum_dirac_dfcoef.utils import debug_print, space_separated_parsing


class STAGE(Enum):
    # STAGE TRANSITION: INIT -> VECTOR_PRINT -> WAIT_END_READING_COEF -> END
    #                                             ↓               ↑
    #                                      WAIT_FIRST_COEF -> READING_COEF
    INIT = auto()
    VECTOR_PRINT = auto()
    WAIT_END_READING_COEF = auto()
    WAIT_FIRST_COEF = auto()
    READING_COEF = auto()
    END = auto()


class PrivecProcessor:
    def __init__(self, dirac_output: TextIOWrapper, functions_info: FunctionsInfo, eigenvalues: Eigenvalues) -> None:
        self.dirac_output = dirac_output
        self.stage = STAGE.INIT
        self.is_electronic = False
        self.eigenvalues = eigenvalues
        self.mo_sym_type = ""
        self.functions_info = functions_info
        self.data_mo = DataMO()
        self.data_all_mo = DataAllMO()
        self.used_atom_info: Dict[str, AtomInfo] = {}
        self.current_atom_info = AtomInfo()

    def read_privec_data(self):
        """Read coefficients from the output file of DIRAC and store them in data_all_mo.

        self.data_all is the final result of this function. You can get all results from this variable except header information.
        """
        self.stage = STAGE.INIT
        for line_str in self.dirac_output:
            words = space_separated_parsing(line_str)

            if self.stage == STAGE.END:
                if args.for_generator:
                    self.fill_non_moltra_range_electronic_eigenvalues()
                break  # End of reading coefficients

            elif self.need_to_skip_this_line(words):
                if self.stage == STAGE.READING_COEF:
                    if self.need_to_create_results_for_current_mo(words):
                        self.add_current_mo_data_to_data_all_mo()
                        self.transition_stage(STAGE.WAIT_END_READING_COEF)

            elif self.stage == STAGE.INIT:
                if self.check_start_vector_print(words):
                    self.transition_stage(STAGE.VECTOR_PRINT)

            elif self.stage == STAGE.VECTOR_PRINT:
                if self.need_to_get_mo_sym_type(words):
                    self.mo_sym_type = words[2]
                    self.transition_stage(STAGE.WAIT_END_READING_COEF)

            elif self.stage == STAGE.WAIT_END_READING_COEF:
                if self.need_to_get_mo_sym_type(words):
                    self.mo_sym_type = words[2]
                elif self.need_to_start_mo_section(words):
                    self.start_mo_section(words)
                    self.transition_stage(STAGE.WAIT_FIRST_COEF)
                elif self.check_end_vector_print(words):
                    self.transition_stage(STAGE.END)

            elif self.stage == STAGE.WAIT_FIRST_COEF:
                if self.is_this_row_for_coefficients(words):
                    self.add_coefficient(line_str)
                    self.transition_stage(STAGE.READING_COEF)

            elif self.stage == STAGE.READING_COEF:
                if self.is_this_row_for_coefficients(words):
                    self.add_coefficient(line_str)

        if not args.no_sort:
            self.data_all_mo.sort_mo_energy()

    def transition_stage(self, new_stage: STAGE) -> None:
        self.stage = new_stage

    def is_this_row_for_coefficients(self, words: List[str]) -> bool:
        # min: 4 coefficients and other words => 5 words
        return True if 5 <= len(words) <= 9 and words[0].isdigit() else False

    def need_to_skip_this_line(self, words: List[str]) -> bool:
        return True if len(words) <= 1 else False

    def need_to_create_results_for_current_mo(self, words: List[str]) -> bool:
        return True if self.stage == STAGE.READING_COEF and len(words) <= 1 else False

    def need_to_get_mo_sym_type(self, words: List[str]) -> bool:
        return True if len(words) == 3 and words[0] == "Fermion" and words[1] == "ircop" else False

    def need_to_start_mo_section(self, words: List[str]) -> bool:
        if words[1] in ("Electronic", "Positronic") and words[2] == "eigenvalue" and "no." in words[3]:
            return True
        return False

    def check_start_vector_print(self, words: List[str]) -> bool:
        # ****************************** Vector print ******************************
        if len(words) < 4:
            return False
        elif words[1] == "Vector" and words[2] == "print":
            return True
        return False

    def check_end_vector_print(self, words: List[str]) -> bool:
        # https://github.com/kohei-noda-qcrg/summarize_dirac_dfcoef_coefficients/issues/7#issuecomment-1377969626
        if len(words) >= 2 and self.stage == STAGE.WAIT_END_READING_COEF:
            return True
        return False

    def get_mo_info(self, eigenvalue_no: int) -> str:
        if args.compress:
            return f"{self.mo_sym_type} {eigenvalue_no}"
        elif self.is_electronic:
            return f"Electronic no. {eigenvalue_no} {self.mo_sym_type}"
        else:
            return f"Positronic no. {eigenvalue_no} {self.mo_sym_type}"

    def start_mo_section(self, words: List[str]) -> None:
        """
        (e.g.)
        words = ["*", "Electronic", "eigenvalue", "no.", "22:", "-2.8417809384721"]
        words = ["*", "Electronic", "eigenvalue", "no.122:", "-2.8417809384721"]
        """

        def set_is_electronic() -> None:
            if words[1] == "Positronic":
                self.is_electronic = False
            elif words[1] == "Electronic":
                self.is_electronic = True
            else:
                msg = f"ERROR: UnKnow MO type, MO_Type={words[1]}"
                raise ValueError(msg)

        def get_eigenvalue_no():
            try:
                eigenvalue_no = int(words[-2][:-1].replace("no.", ""))
            except ValueError:
                # If *** is printed, we have no information about what number this MO is.
                # Therefore, we assume that eigenvalue_no is the next number after prev_eigenvalue_no.
                prev_eigenvalue_no = self.data_mo.eigenvalue_no  # prev_electron is the number of electrons of the previous MO
                eigenvalue_no = prev_eigenvalue_no + 1
            return eigenvalue_no

        set_is_electronic()
        eigenvalue_no = get_eigenvalue_no()
        mo_energy = float(words[-1])
        mo_info = self.get_mo_info(eigenvalue_no)

        # Here is the start point of reading coefficients of the current MO
        self.data_mo.reset()  # reset data_mo because we need to delete data_mo of the previous MO
        self.data_mo.eigenvalue_no = eigenvalue_no
        self.data_mo.mo_energy = mo_energy
        self.data_mo.mo_info = mo_info
        self.used_atom_info.clear()  # reset used_atom_info because we need to delete used_atom_info of the previous MO

    def add_coefficient(self, line_str: str) -> None:
        component_func = "large" if line_str[10] == "L" else ("small" if line_str[10] == "S" else "")  # CLS
        symmetry_label = line_str[12:15].strip()  # REP (e.g. "Ag "), symmetry_label="Ag"
        atom_label = line_str[15:18].strip()  # NAMN (e.g. "Cm "), atom_labe="Cm"
        gto_type = line_str[18:22].strip()  # GTOTYP (e.g. "s   "), gto_type="s"
        label = symmetry_label + atom_label

        if self.current_atom_info.count_remaining_functions() == 0 or label != self.current_atom_info.label:
            # First, we need to read information about the current atom.
            if label not in self.used_atom_info:
                # It is the first time to read information about the current atom.
                cur_atom_start_idx = 1
            else:
                # It is not the first time to read information about the current atom.
                # So we need to read information about the previous atom from used_atom_info.
                # current start_idx = previous start_idx + previous mul
                cur_atom_start_idx = self.used_atom_info[label].start_idx + self.used_atom_info[label].mul
            # Validate start_idx
            if cur_atom_start_idx not in self.functions_info[component_func][symmetry_label][atom_label]:
                msg = f"start_idx={cur_atom_start_idx} is not found in functions_info[{component_func}][{symmetry_label}][{atom_label}]"
                raise Exception(msg)
            # We can get information about the current atom from functions_info with start_idx.
            self.current_atom_info = copy.deepcopy(self.functions_info[component_func][symmetry_label][atom_label][cur_atom_start_idx])
            # Update used_atom_info with current_atom_info
            self.used_atom_info[label] = copy.deepcopy(self.current_atom_info)

        self.current_atom_info.decrement_function(gto_type)
        self.data_mo.add_coefficient(get_coefficient(line_str, self.functions_info, self.current_atom_info.start_idx))

    def add_current_mo_data_to_data_all_mo(self) -> None:
        self.data_mo.fileter_coefficients_by_threshold()
        if self.is_electronic:
            self.data_all_mo.electronic.append(copy.deepcopy(self.data_mo))
            cur_sym = self.mo_sym_type
            if args.for_generator:
                self.eigenvalues.energies_used[cur_sym][self.data_mo.eigenvalue_no] = True
        else:
            self.data_all_mo.positronic.append(copy.deepcopy(self.data_mo))
        debug_print(f"End of reading {self.data_mo.eigenvalue_no}th MO")

    def fill_non_moltra_range_electronic_eigenvalues(self):
        self.is_electronic = True
        for sym_type_key, val in self.eigenvalues.energies_used.items():
            self.mo_sym_type = sym_type_key
            for eigenvalue_no, is_found in val.items():
                if not is_found:
                    self.data_mo.reset()
                    self.data_mo.eigenvalue_no = eigenvalue_no
                    self.data_mo.mo_info = self.get_mo_info(eigenvalue_no)
                    self.data_mo.mo_energy = self.eigenvalues.energies[sym_type_key][eigenvalue_no - 1]
                    self.data_all_mo.electronic.append(copy.deepcopy(self.data_mo))



import re
from collections import OrderedDict
from io import TextIOWrapper
from typing import ClassVar, Dict, List
from typing import OrderedDict as ODict

from sum_dirac_dfcoef.utils import (
    debug_print,
    delete_dirac_input_comment_out,
    is_dirac_input_keyword,
    is_dirac_input_line_should_be_skipped,
    is_dirac_input_section_one_star,
    is_end_dirac_input_field,
    is_start_dirac_input_field,
    space_separated_parsing,
    space_separated_parsing_upper,
)


# type definition eigenvalues.shell_num
# type eigenvalues = {
#     "E1g": {
#         "closed": int
#         "open": int
#         "virtual": int
#     },
#     "E1u": {
#         "closed": int
#         "open": int
#         "virtual": int
#     },
# }
class Eigenvalues:
    shell_num: ClassVar[ODict[str, Dict[str, int]]] = OrderedDict()
    energies: ClassVar[ODict[str, List[float]]] = OrderedDict()
    energies_used: ClassVar[ODict[str, Dict[int, bool]]] = OrderedDict()

    def setdefault(self, key: str):
        self.shell_num.setdefault(key, {"closed": 0, "open": 0, "virtual": 0, "negative": 0, "positronic": 0})
        self.energies.setdefault(key, [])
        self.energies_used.setdefault(key, {})

    def get_electronic_spinor_num(self, symmetry_type: str) -> int:
        return self.shell_num[symmetry_type]["closed"] + self.shell_num[symmetry_type]["open"] + self.shell_num[symmetry_type]["virtual"]

    def get_eigenvalues(self, dirac_output: TextIOWrapper):
        def is_end_of_read(line) -> bool:
            if "HOMO - LUMO" in line:
                return True
            return False

        def is_eigenvalue_type_written(words: List[str]) -> bool:
            # closed shell: https://gitlab.com/dirac/dirac/-/blob/364663fd2bcc419e41ad01703fd782889435b576/src/dirac/dirout.F#L1043
            # open shell: https://gitlab.com/dirac/dirac/-/blob/364663fd2bcc419e41ad01703fd782889435b576/src/dirac/dirout.F#L1053
            # virtual eigenvalues: https://gitlab.com/dirac/dirac/-/blob/364663fd2bcc419e41ad01703fd782889435b576/src/dirac/dirout.F#L1064
            # negative energy eigenvalues (only atom or linear molecule case): https://gitlab.com/dirac/dirac/-/blob/364663fd2bcc419e41ad01703fd782889435b576/src/dirac/dirout.F#L1156
            # positronic eigenvalues (not atom and linear molecule case): https://gitlab.com/dirac/dirac/-/blob/364663fd2bcc419e41ad01703fd782889435b576/src/dirac/dirout.F#L1073
            if "*" == words[0] and "Closed" == words[1] and "shell," == words[2]:
                return True
            elif "*" == words[0] and "Open" == words[1] and "shell" == words[2]:
                return True
            elif "*" == words[0] and "Virtual" == words[1] and "eigenvalues," == words[2]:
                return True
            elif "*" == words[0] and "Negative" == words[1] and "energy" == words[2] and "eigenvalues," == words[3]:
                return True
            elif "*" == words[0] and "Positronic" == words[1] and "eigenvalues," == words[2]:
                return True
            return False

        def get_current_eigenvalue_type(words: List[str]) -> str:
            # words[0] = '*', words[1] = "Closed" or "Open" or "Virtual" or "Negative" or "Positronic"
            current_eigenvalue_type = words[1].lower()
            return current_eigenvalue_type

        def get_symmetry_type_standard(words: List[str]) -> str:
            current_symmetry_type = words[3]
            return current_symmetry_type

        def get_symmetry_type_supersym(words: List[str]) -> str:
            # https://gitlab.com/dirac/dirac/-/blob/364663fd2bcc419e41ad01703fd782889435b576/src/dirac/dirout.F#L1097-1105
            # FORMAT '(/A,I4,4A,I2,...)'
            # DATA "* Block",ISUB,' in ',FREP(IFSYM),":  ",...
            # ISUB might be **** if ISUB > 9999 or ISUB < -999 because of the format
            # Therefore, find 'in' word list and get FREP(IFSYM) from the word list
            # FREP(IFSYM) is a symmetry type
            idx = words.index("in")
            current_symmetry_type = words[idx + 1][: len(words[idx + 1]) - 1]
            return current_symmetry_type

        scf_cycle = False
        eigenvalues_header = False
        print_type = ""  # "standard" or "supersymmetry"
        current_eigenvalue_type = ""  # "closed" or "open" or "virtual"
        current_symmetry_type = ""  # "E1g" or "E1u" or "E1" ...

        for line in dirac_output:
            words: List[str] = space_separated_parsing(line)

            if len(words) == 0:
                continue

            if "SCF - CYCLE" in line:
                scf_cycle = True
                continue

            if scf_cycle and not eigenvalues_header:
                if "Eigenvalues" == words[0]:
                    eigenvalues_header = True
                continue

            if print_type == "":  # search print type (standard or supersymmetry)
                if "*" == words[0] and "Fermion" in words[1] and "symmetry" in words[2]:
                    print_type = "standard"
                    current_symmetry_type = get_symmetry_type_standard(words)
                    self.setdefault(current_symmetry_type)
                elif "* Block" in line:
                    print_type = "supersymmetry"
                    current_symmetry_type = get_symmetry_type_supersym(words)
                    self.setdefault(current_symmetry_type)
                continue

            if print_type == "standard" and "*" == words[0] and "Fermion" in words[1] and "symmetry" in words[2]:
                current_symmetry_type = get_symmetry_type_standard(words)
                self.setdefault(current_symmetry_type)
            elif print_type == "supersymmetry" and "* Block" in line:
                current_symmetry_type = get_symmetry_type_supersym(words)
                self.setdefault(current_symmetry_type)
            elif is_eigenvalue_type_written(words):
                current_eigenvalue_type = get_current_eigenvalue_type(words)
            elif is_end_of_read(line):
                break
            else:
                start_idx = 0
                while True:
                    # e.g. -775.202926514  ( 2) => -775.202926514
                    regex = r"[-]?[0-9]+\.?[0-9]+"
                    match = re.search(regex, line[start_idx:])
                    if match is None:
                        break
                    val = float(match.group())

                    # e.g. -775.202926514  ( 2) => 2
                    regex = r"\([ ]*[0-9]+\)"
                    match = re.search(regex, line[start_idx:])
                    if match is None:
                        break
                    # match.group() == ( 2) => [1 : len(match.group()) - 1] == 2
                    num = int(match.group()[1 : len(match.group()) - 1])
                    self.shell_num[current_symmetry_type][current_eigenvalue_type] += num
                    for _ in range(0, num, 2):
                        self.energies[current_symmetry_type].append(val)
                        self.energies_used[current_symmetry_type][len(self.energies[current_symmetry_type])] = False
                    start_idx += match.end()

        for key in self.energies.keys():
            self.energies[key].sort()
        debug_print(f"eigenvalues: {self}")

    def validate_eigpri_option(self, dirac_output: TextIOWrapper):
        """Validate the .EIGPRI option in the DIRAC input file,
        if is not set, it is a valid input
        because only the positive energy eigenvalues are printed as default.

        Args:
            dirac_output (TextIOWrapper): _description_
        """

        is_reach_input_field: bool = False
        is_scf_section: bool = False
        is_scf_detail_section: bool = False
        is_next_line_eigpri: bool = False
        for line in dirac_output:
            no_comment_out_line = delete_dirac_input_comment_out(line)
            words = space_separated_parsing_upper(no_comment_out_line)
            if is_dirac_input_line_should_be_skipped(words):
                continue

            if is_start_dirac_input_field(no_comment_out_line):
                is_reach_input_field = True
                continue

            if is_end_dirac_input_field(no_comment_out_line):
                break

            if is_reach_input_field:
                if is_dirac_input_keyword(words[0]):
                    if ".SCF" in words[0]:
                        is_scf_section = True
                        continue

            if is_scf_section:
                if is_dirac_input_section_one_star(words[0]):
                    if "*SCF" in words[0]:
                        is_scf_detail_section = True
                        continue
                    else:
                        is_scf_detail_section = False
                        continue

            if is_scf_detail_section:
                if is_dirac_input_keyword(words[0]):
                    if ".EIGPRI" in words[0]:
                        is_next_line_eigpri = True
                        continue
                    else:
                        is_next_line_eigpri = False
                        continue

            if is_next_line_eigpri:
                # https://diracprogram.org/doc/master/manual/wave_function/scf.html#eigpri
                if len(words) == 2 and words[0].isdigit() and words[1].isdigit():
                    if int(words[0]) == 0:  # positive energy eigenvalues are not printed
                        msg = f"\nYour .EIGPRI option in your DIRAC input file is invalid!\n\
.EIGPRI\n\
{line}\n\
We cannot get the eigenvalues with your .EIGPRI option.\n\
If you want to use this output file with this program, you must use --no-scf option to skip reading eigenvalues information.\n\
But you cannot use the output using --no-scf option to dcaspt2_input_generator program.\n\
If you want to get the eigenvalues information, please refer the .EIGPRI option in the manual of DIRAC.\n\
https://diracprogram.org/doc/master/manual/wave_function/scf.html#eigpri\n\
You must enable to print out the positive eigenvalues energy.\n"
                        raise ValueError(msg)



from typing import ClassVar, Dict, List

from sum_dirac_dfcoef.args import args
from sum_dirac_dfcoef.coefficient import Coefficient


class DataMO:
    norm_const_sum: float = 0.0
    mo_energy: float = 0.0
    mo_info: str = ""
    eigenvalue_no: int = 0
    coef_dict: ClassVar[Dict[str, Coefficient]] = {}
    coef_list: ClassVar[List[Coefficient]] = []

    def __repr__(self) -> str:
        return f"norm_const_sum: {self.norm_const_sum}, coef_dict: {self.coef_dict}"

    def add_coefficient(self, coef: Coefficient) -> None:
        key = coef.function_label + str(coef.start_idx)
        if key in self.coef_dict:
            self.coef_dict[key].coefficient += coef.coefficient
        else:
            self.coef_dict[key] = coef
        self.norm_const_sum += coef.coefficient * coef.multiplication

    def reset(self):
        self.norm_const_sum = 0.0
        self.mo_energy = 0.0
        self.mo_info = ""
        self.eigenvalue_no = 0
        self.coef_dict.clear()
        self.coef_list.clear()

    def fileter_coefficients_by_threshold(self) -> None:
        self.coef_list = [coef for coef in self.coef_dict.values() if abs(coef.coefficient / self.norm_const_sum * 100) >= args.threshold]
        self.coef_list.sort(key=lambda coef: coef.coefficient, reverse=True)


class DataAllMO:
    electronic: ClassVar[List[DataMO]] = []
    positronic: ClassVar[List[DataMO]] = []

    def __repr__(self) -> str:
        return f"electronic: {self.electronic}, positronic: {self.positronic}"

    def sort_mo_energy(self) -> None:
        self.electronic.sort(key=lambda mo: mo.mo_energy)
        self.positronic.sort(key=lambda mo: mo.mo_energy)
